#include "stdafx.h"
#include "3dsloader.h"

// supported chunks:
#define CH_MAGIC      0x4d4d
#define  CH_VERS      0x0002
#define  CH_MDATA     0x3d3d
#define   CH_MVERS    0x3d3e
#define   CH_SCALE    0x0100
#define   CH_MATER    0xafff
#define   CH_OBJEC    0x4000
#define    CH_TROBJ   0x4100
#define     CH_POINT  0x4110
#define     CH_TEXV   0x4140
#define     CH_FACE   0x4120
#define      CH_SMOOG 0x4150
#define      CH_MATG  0x4130
#define     CH_MATRI  0x4160

C3DSLoader::C3DSLoader()
{
  m_nObjects = 0;

  m_data = NULL;
  m_datalen = 0;
}

C3DSLoader::~C3DSLoader()
{
  for(int i=0; i<m_nObjects; i++)
  {
    if(m_pObjects[i].vertices != NULL)
      delete []m_pObjects[i].vertices;

    if(m_pObjects[i].faces != NULL)
      delete []m_pObjects[i].faces;
  }

  m_nObjects = 0;

  assert(m_data == NULL);
}

unsigned short C3DSLoader::ReadShort(unsigned int ofs)
{
  assert(ofs + 2 <= m_datalen);

  return m_data[ofs] + 0x100*m_data[ofs+1];
}

unsigned int C3DSLoader::ReadInt(unsigned int ofs)
{
  assert(ofs + 4 <= m_datalen);

  return m_data[ofs] + 0x100*m_data[ofs+1] + 0x10000*m_data[ofs+2] + 0x1000000*m_data[ofs+3];
}

float C3DSLoader::ReadFloat(unsigned int ofs)
{
  float f;
  unsigned char *fc = (unsigned char *)&f;
  assert(ofs + 4 <= m_datalen);

  for(int i=0; i<4; i++)
  {
#ifdef SPARC
    fc[i] = m_data[ofs+3-i];
#else
    fc[i] = m_data[ofs+i];
#endif
  }

  return f;
}

bool C3DSLoader::SkipName(load3ds_chunk_t *chunk)
{
  unsigned int o = chunk->dataofs;
  while(o < m_datalen && m_data[o] != 0)
  {
    o++;
  }

  if(o >= m_datalen)
    return false;

  int dataend = chunk->dataofs + chunk->datalen;

  chunk->dataofs = o+1;
  chunk->datalen = dataend - chunk->dataofs;

  if(chunk->datalen < 0)
  {
    assert(false);
    return false;
  }

  return true;
}

bool C3DSLoader::ReadChunk(load3ds_chunk_t *chunk, unsigned int ofs, load3ds_chunk_t *parentchunk)
{
  unsigned int end_of_block = m_datalen;

  if(parentchunk != NULL)
    end_of_block = parentchunk->dataofs + parentchunk->datalen;

  if(ofs + 6 > end_of_block)
    return false;

  chunk->id = ReadShort(ofs);
  chunk->len = ReadInt(ofs+2);

  if(ofs + chunk->len > end_of_block)
    return false;

  chunk->dataofs = ofs+6;
  chunk->datalen = chunk->len-6;

  if(HasName(chunk->id))
  {
    if(!SkipName(chunk))
      return false;
  }

  chunk->data = &m_data[chunk->dataofs];

  chunk->params = NULL;
  chunk->paramsofs = -1;
  chunk->paramslen = 0;

  chunk->next = NULL;
  chunk->child = NULL;

  return true;
}

bool C3DSLoader::HasChild(unsigned short pid, unsigned short id)
{
  int i=0;
  // {parent, child}
  unsigned short ids[][2] = {
  {CH_MAGIC, CH_VERS}, 
  {CH_MAGIC, CH_MDATA}, 
  {CH_MDATA, CH_MVERS}, 
  {CH_MDATA, CH_SCALE}, 
  {CH_MDATA, CH_MATER}, 
  {CH_MDATA, CH_OBJEC}, 
  {CH_OBJEC, CH_TROBJ}, 
  {CH_TROBJ, CH_POINT},
  {CH_TROBJ, CH_TEXV},
  {CH_TROBJ, CH_FACE},
  {CH_TROBJ, CH_MATRI},
  {CH_FACE, CH_SMOOG},
  {CH_FACE, CH_MATG},
  {0,0}};

  while(ids[i][0] != 0)
  {
    if(ids[i][0] == pid && ids[i][1] == id)
      return true;
    i++;
  }

  return false;
}

bool C3DSLoader::HasName(unsigned short id)
{
  int i=0;
  unsigned short ids[] = {{CH_OBJEC},
  {0}};

  while(ids[i] != 0)
  {
    if(ids[i] == id)
      return true;
    i++;
  }

  return false;
}

unsigned short C3DSLoader::ReadPossibleId(unsigned int ofs)
{
  if(ofs < 0 || ofs+2 > m_datalen)
    return 0x0000; //return dummy id

  return ReadShort(ofs);
}

bool C3DSLoader::ReadChunkTree(load3ds_chunk_t *rootchunk)
{
  load3ds_chunk_t *ch;
  load3ds_chunk_t **predecessor = &rootchunk->child;
  unsigned int o = rootchunk->dataofs;

  if(rootchunk->datalen >= 6) // has possibly subblock
  {
    do
    {
      unsigned short possible_id = ReadPossibleId(o);

      if(HasChild(rootchunk->id, possible_id))
      {
        ch = new load3ds_chunk_t;
        assert(ch);

        if(!ReadChunk(ch, o, NULL))
        {
          assert(false);
          return false;
        }

        if(!ReadChunkTree(ch))
          return false;
      }
      else
      {
        rootchunk->paramsofs = o;
        rootchunk->params = &m_data[o];
        rootchunk->paramslen = rootchunk->dataofs + rootchunk->datalen - o;
        if(rootchunk->paramslen < 0)
        {
          assert(false);
          return false;
        }
        return true;
      }

      o = ch->dataofs + ch->datalen;

      ch->next = NULL;
      *predecessor = ch;
      predecessor = &ch->next;
    }
    while(o < rootchunk->dataofs + rootchunk->datalen);

    rootchunk->paramsofs = -1;
    rootchunk->params = NULL;
    rootchunk->paramslen = 0; // no space for data left
  }
  else
  {
    // all data used for parameters
    rootchunk->paramsofs = rootchunk->dataofs;
    rootchunk->params = rootchunk->data;
    rootchunk->paramslen = rootchunk->datalen;
  }

  return true;
}

load3ds_chunk_t *C3DSLoader::FindSubChunk(load3ds_chunk_t *parent, unsigned short id, int cnt)
{
  load3ds_chunk_t *c;
  
  c = parent->child;

  while(c != NULL && cnt >= 0)
  {
    if(c->id == id)
    {
      if(cnt == 0)
        return c;

      cnt--;
    }

     c = c->next;
  }

  return NULL;
}

bool C3DSLoader::Load3DS(const char *filename)
{
  unsigned int j;

  FILE *file = fopen(filename, "rb");
  if(file == NULL)
    return false;

  fseek(file, 0, 2);
  m_datalen = ftell(file);
  fseek(file, 0, 0);

  m_data = new unsigned char[m_datalen];

  if(m_data == NULL)
  {
    fclose(file);
    return false;
  }

  fread(m_data, m_datalen, 1, file);

  fclose(file);

  load3ds_chunk_t root;
  if(!ReadChunk(&root, 0, NULL))
    return false;

  if(!ReadChunkTree(&root))
    return false;

  load3ds_chunk_t *meshdata = FindSubChunk(&root, CH_MDATA);
  if(meshdata == NULL)
    return false;

  m_nObjects = 0;

  while(FindSubChunk(meshdata, CH_OBJEC, m_nObjects))
  {
    m_nObjects ++;
  }

  if(m_nObjects > MAX_3DS_OBJECTS)
    return false;

  for(int i=0; i<m_nObjects; i++)
  {
    load3ds_chunk_t *obj  = FindSubChunk(meshdata, CH_OBJEC, i);
    load3ds_chunk_t *mesh = FindSubChunk(obj, CH_TROBJ);
    load3ds_chunk_t *vtx = FindSubChunk(mesh, CH_POINT);

    unsigned int nvtx = ReadShort(vtx->paramsofs);

    if(vtx->paramslen != 2 + nvtx*4*3)
      return false;

    m_pObjects[i].nvertices = nvtx;
    m_pObjects[i].vertices = new load3ds_vertex_t[nvtx];

    for(j=0; j<nvtx; j++)
    {
      m_pObjects[i].vertices[j].vlocal.x = ReadFloat(vtx->paramsofs + 2 + 4*(j*3 + 0));
      m_pObjects[i].vertices[j].vlocal.z = ReadFloat(vtx->paramsofs + 2 + 4*(j*3 + 1));
      m_pObjects[i].vertices[j].vlocal.y = ReadFloat(vtx->paramsofs + 2 + 4*(j*3 + 2));

      m_pObjects[i].vertices[j].uv_set = false;
    }

    load3ds_chunk_t *faces = FindSubChunk(mesh, CH_FACE);

    unsigned int nfaces = ReadShort(faces->paramsofs);

    if(faces->paramslen < 2 + nfaces*4*2)
      return false;

    m_pObjects[i].nfaces = nfaces;
    m_pObjects[i].faces = new load3ds_face_t[nfaces];

    for(unsigned int f=0; f<nfaces; f++)
    {
      m_pObjects[i].faces[f].v_index[0] = ReadShort(faces->paramsofs + 2 + 2*(f*4 + 0));
      m_pObjects[i].faces[f].v_index[1] = ReadShort(faces->paramsofs + 2 + 2*(f*4 + 1));
      m_pObjects[i].faces[f].v_index[2] = ReadShort(faces->paramsofs + 2 + 2*(f*4 + 2));
    }

    load3ds_chunk_t *texv = FindSubChunk(mesh, CH_TEXV);

    if(texv != NULL)
    {
      unsigned int nvtx_texv = ReadShort(texv->paramsofs);

      if(nvtx_texv == nvtx && texv->paramslen == 2 + nvtx*4*2)
      {
        for(j=0; j<nvtx; j++)
        {
          m_pObjects[i].vertices[j].uv_set = true;
          m_pObjects[i].vertices[j].uv.u = ReadFloat(texv->paramsofs + 2 + 4*(j*2+0));
          m_pObjects[i].vertices[j].uv.v = ReadFloat(texv->paramsofs + 2 + 4*(j*2+1));
        }
      }
    }
  }

  delete []m_data;
  m_data = NULL;

  return true;
}
