#include "stdafx.h"
#include "gllib.h"

#ifdef SPARC
  #define DISABLE_EXTENSIONS // no extensions if compiled for SPARC
#endif

#ifndef DISABLE_EXTENSIONS
  #ifdef WIN32
    #define MyGetProcAddress wglGetProcAddress
  #else
    #define GLX_GLXEXT_PROTOTYPES  
    #include "GL/glx.h"

    #define MyGetProcAddress glXGetProcAddressARB 
  #endif
#endif

#define R(x) (((x)>>16)&255)
#define G(x) (((x)>>8)&255)
#define B(x) (((x)>>0)&255)

#define FR(x) (R(x)/255.0f)
#define FG(x) (G(x)/255.0f)
#define FB(x) (B(x)/255.0f)

GLLib::GLLib()
{
  glActiveTextureARB = NULL;
  glClientActiveTextureARB = NULL;
  glMultiTexCoord2fvARB = NULL;

  m_lasterror = GL_NO_ERROR;

  m_cam_matrix_pushed = false;

  for(int i=0; i<MAX_TEXTURE_STAGES; i++)
    m_texstage_enabled[i] = false;

  m_transparency = 0.5f;

  InitGL();
}

bool GLLib::GetARBExtensions()
{
#ifndef DISABLE_EXTENSIONS
  char *strExtensions = (char*)glGetString(GL_EXTENSIONS);

  if(strstr(strExtensions, "GL_ARB_multitexture") == 0) 
  {
    // multitexturing (GL_ARB_multitexture) not supported
    return false;
  }
  
#ifndef WIN32
  glActiveTextureARB       = (PFNGLACTIVETEXTUREARBPROC)MyGetProcAddress((const GLubyte*)"glActiveTextureARB");
  glClientActiveTextureARB = (PFNGLCLIENTACTIVETEXTUREARBPROC)MyGetProcAddress((const GLubyte*)"glClientActiveTextureARB");
  glMultiTexCoord2fvARB    = (PFNGLMULTITEXCOORD2FVARBPROC)MyGetProcAddress((const GLubyte*)"glMultiTexCoord2fvARB");
#else
  glActiveTextureARB       = (PFNGLACTIVETEXTUREARBPROC)MyGetProcAddress("glActiveTextureARB");
  glClientActiveTextureARB = (PFNGLCLIENTACTIVETEXTUREARBPROC)MyGetProcAddress("glClientActiveTextureARB");
  glMultiTexCoord2fvARB    = (PFNGLMULTITEXCOORD2FVARBPROC)MyGetProcAddress("glMultiTexCoord2fvARB");
#endif

  if(glActiveTextureARB == NULL)
  {
    // multitexturing (glActiveTextureARB) not supported
    return false;
  }

  if(glClientActiveTextureARB == NULL)
  {
    // multitexturing (glClientActiveTextureARB) not supported
    return false;
  }

  if(glMultiTexCoord2fvARB == NULL)
  {
    // multitexturing (glMultiTexCoord2fvARB) not supported
    return false;
  }
#endif

  return true;
}

void GLLib::InitGL()
{
  if(!GetARBExtensions())
  {
    assert(0);
  }

  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity();
  
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);
  glPolygonMode(GL_FRONT, GL_FILL);

  glDepthFunc(GL_LEQUAL);

  if(!CheckForError())
    return;

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);

  static float ambient[]  = {0.9f, 0.9f, 0.9f, 0.9f};
  static float diffuse[]  = {1.0f, 1.0f, 1.0f, 1.0f};
  static float specular[] = {1.0f, 1.0f, 1.0f, 1.0f};
  
  glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
  glLightfv(GL_LIGHT0, GL_SPECULAR, specular);

  if(!CheckForError())
    return;
}

GLLib::~GLLib()
{
}

bool GLLib::CheckForError()
{
  int e = glGetError();
  if(e != GL_NO_ERROR)
  {
    assert(0);
    m_lasterror = e;
    return false;
  }

  return true;
}

int GLLib::GetLastError()
{
  return m_lasterror;
}

bool GLLib::SetWindowSize(int w, int h)
{
  if(!CheckForError())
    return false;

  glViewport(0, 0, w, h);

  if(!CheckForError())
    return false;

  return true;
}

bool GLLib::StartRender(GLLIB_COLOR background_color, BOOL clear_screen)
{
  if(!CheckForError())
    return false;

  glClearColor(FR(background_color), FG(background_color), FB(background_color), 0);

  int clearflags = GL_DEPTH_BUFFER_BIT;

  if(clear_screen)
    clearflags |= GL_COLOR_BUFFER_BIT;

  glClear(clearflags);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  if(!CheckForError())
    return false;

  return true;
}

bool GLLib::EndRender()
{
  if(!CheckForError())
    return false;

  // NOTE: buffers must be swapped by caller
  
  return true;
}

bool GLLib::RenderPoint(GLLIB_vtx *vtx, float size)
{
  if(!CheckForError())
    return false;

  glPointSize(size);

  DisableTexture(0);  

  glBegin(GL_POINTS);

  glVertex3fv((float *)vtx);

  glEnd();

  if(!CheckForError())
    return false;

  return true;
}

bool GLLib::RenderLine(GLLIB_vtx *v1, GLLIB_vtx *v2)
{
  if(!CheckForError())
    return false;

  DisableTexture(0);
  glDisable(GL_TEXTURE_2D);
  
  glLineWidth(2.0f);

  glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
  glDisable(GL_CULL_FACE);

  glBegin(GL_TRIANGLES);

  glVertex3fv((float *)v1);
  glVertex3fv((float *)v2);
  glVertex3fv((float *)v2);

  glEnd();
   
  glEnable(GL_CULL_FACE);
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  glEnable(GL_TEXTURE_2D);

  if(!CheckForError())
    return false;

  return true;
}

bool GLLib::RenderFace(GLLIB_vtx *vtxlist, GLLIB_uvvtx *uvlist, GLLIB_vtx *normallist)
{
  if(!CheckForError())
    return false;

  if(!m_texstage_enabled[0]) // disable texture coordinates if no texture set
    uvlist = NULL;

  glBegin(GL_TRIANGLES);

  for(int i=0; i<3; i++)
  {
    if(uvlist != NULL)
    {
#ifdef DISABLE_EXTENSIONS
      glTexCoord2fv((float *)&uvlist[i]);  
#else
      glMultiTexCoord2fvARB(GL_TEXTURE0_ARB, (float *)&uvlist[i]);
#endif
    }

    if(normallist != NULL)
      glNormal3fv((float *)&normallist[i]);

    glVertex3fv((float *)&vtxlist[i]);
  }

  glEnd();
  
  if(!CheckForError())
    return false;

  return true;
}

bool GLLib::RenderFaces(GLLIB_vtx *vtxlist, GLLIB_uvvtx *uv1list, GLLIB_vtx *normlist, int *facelist, int n_faces)
{
  if(!CheckForError())
    return false;

  if(!m_texstage_enabled[0]) // disable texture coordinates if no texture set
    uv1list = NULL;

  glVertexPointer(3, GL_FLOAT, 0, vtxlist);
  glEnableClientState(GL_VERTEX_ARRAY);

  if(normlist != NULL)
  {
    glNormalPointer(GL_FLOAT, 0, normlist);
    glEnableClientState(GL_NORMAL_ARRAY);
  }

  if(uv1list != NULL)
  {
#ifndef DISABLE_EXTENSIONS
    glClientActiveTextureARB(GL_TEXTURE0_ARB);
#endif
    glTexCoordPointer(2, GL_FLOAT, 0, uv1list);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  }

  // rendering

  glDrawElements(GL_TRIANGLES, n_faces*3, GL_UNSIGNED_INT, facelist);

  glDisableClientState(GL_VERTEX_ARRAY);

  if(normlist != NULL)
  {
    glDisableClientState(GL_NORMAL_ARRAY);
  }

  if(uv1list != NULL)
  {
#ifndef DISABLE_EXTENSIONS
    glClientActiveTextureARB(GL_TEXTURE0_ARB);
#endif
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  }

  if(!CheckForError())
    return false;

  return true;
}

bool GLLib::RenderFacesWireframe(GLLIB_vtx *vtxlist, int *facelist, int n_faces)
{
  if(!CheckForError())
    return false;

  DisableTexture(0);
  glDisable(GL_TEXTURE_2D);
  
  glLineWidth(2.0f);

  glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
  glDisable(GL_CULL_FACE);

  glVertexPointer(3, GL_FLOAT, 0, vtxlist);
  glEnableClientState(GL_VERTEX_ARRAY);

  glDrawElements(GL_TRIANGLES, n_faces*3, GL_UNSIGNED_INT, facelist);

  glDisableClientState(GL_VERTEX_ARRAY);

  glEnable(GL_CULL_FACE);
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  glEnable(GL_TEXTURE_2D);

  if(!CheckForError())
    return false;

  return true;
}

bool GLLib::SetLights()
{
  if(!CheckForError())
    return false;

  // currently, we've only got one static light:
  static float position[] = {0.1f, 0.4f, -1.0f, 0.0f};
  glLightfv(GL_LIGHT0, GL_POSITION, position);

  if(!CheckForError())
    return false;

  return true;
}

bool GLLib::SetCamera(GLLIB_vtx &pos, GLLIB_vtx &lookat, GLLIB_vtx &up)
{
  if(!CheckForError())
    return false;

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();   
  gluPerspective(60.0f, 320.0 / 240.0, GLLIB_NEAR_Z_CLIP, GLLIB_FAR_Z_CLIP);

  glMatrixMode(GL_MODELVIEW);

  if(m_cam_matrix_pushed) // if there's already a cam matrix on the stack
    glPopMatrix();

  glLoadIdentity();
  gluLookAt(pos.x, pos.y, pos.z, lookat.x, lookat.y, lookat.z, up.x, up.y, up.z);
 
  glPushMatrix(); // save cam matrix

  m_cam_matrix_pushed = TRUE;

  if(!SetLights())
    return false;

  if(!CheckForError())
    return false;

  return true;
}

unsigned GLLib::CreateTexture(int *data, int w, int h)
{
  unsigned texid;

  if(!CheckForError())
    return 0;

  glGenTextures(1, &texid);

  if(!CheckForError())
    return 0;

  glBindTexture(GL_TEXTURE_2D, texid);
  glTexImage2D(GL_TEXTURE_2D, 0, 3, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);

  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_LINEAR);

  return texid;
}

bool GLLib::DelTexture(unsigned texid)
{
  int tempdata[16*16];

  if(!CheckForError())
    return false;

  // set texture to a 16x16 pixels dummy texture to save memory
  glBindTexture(GL_TEXTURE_2D, texid);
  glTexImage2D(GL_TEXTURE_2D, 0, 3, 16, 16, 0, GL_RGBA, GL_UNSIGNED_BYTE, tempdata);

  return true;
}

bool GLLib::SetCurrentTexture(int stage, unsigned texid)
{
  int gl_stage;

  if(!CheckForError())
    return false;

  assert(stage >= 0 && stage < MAX_TEXTURE_STAGES);

  m_texstage_enabled[stage] = true;

  if(stage == 0)
    gl_stage = GL_TEXTURE0_ARB;
  else
  if(stage == 1)
    gl_stage = GL_TEXTURE1_ARB;
  else
  {
    assert(0);
    return false;
  }

#ifndef DISABLE_EXTENSIONS
  glActiveTextureARB(gl_stage);
#endif

  glEnable(GL_TEXTURE_2D);
  glMatrixMode(GL_TEXTURE);
  glLoadIdentity();

  if(!CheckForError())
    return false;

#ifndef DISABLE_EXTENSIONS
  glActiveTextureARB(gl_stage);
#endif

  glBindTexture(GL_TEXTURE_2D, texid);
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

  if(!CheckForError())
    return false;

  if(stage == 1)
  {
#ifdef DISABLE_EXTENSIONS
    assert(0);
    return false; // this isn't allowed without extensions!
#endif

    // add second texture to first one
    float mask_cmultip[4] = {1.0f, 1.0f, 1.0f, 0.0f};
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);
    glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, mask_cmultip);
  }

  if(!CheckForError())
    return false;

  return true;
}

bool GLLib::SetMaterial(GLLIB_COLOR ambient, GLLIB_COLOR diffuse, GLLIB_COLOR specular, float shininess)
{
  if(!CheckForError())
    return false;

  assert(shininess >= 0.0f && shininess <= 1.0f);

  GLfloat mambient[] = {FR(ambient), FG(ambient), FB(ambient), m_transparency};
  GLfloat mdiffuse[] = {FR(diffuse), FG(diffuse), FB(diffuse), m_transparency};
  GLfloat mspecular[] = {FR(specular), FG(specular), FB(specular), m_transparency};
  GLfloat mshininess = shininess*128;

  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, mambient);
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, mdiffuse);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mspecular);
  glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, mshininess);

  if(!CheckForError())
    return false;

  return true;
}

bool GLLib::DisableTexture(int stage)
{
  int gl_stage;

  assert(stage >= 0 && stage < MAX_TEXTURE_STAGES);

  if(!CheckForError())
    return false;

  m_texstage_enabled[stage] = false;

  if(stage == 0)
    gl_stage = GL_TEXTURE0_ARB;
  else
  if(stage == 1)
    gl_stage = GL_TEXTURE1_ARB;
  else
  {
    assert(0);
    return false;
  }

#ifndef DISABLE_EXTENSIONS
  glActiveTextureARB(gl_stage);
#endif
  glDisable(GL_TEXTURE_2D);

  if(!CheckForError())
    return false;

  return true;
}

bool GLLib::DisableTransparency()
{
  if(!CheckForError())
    return false;

  glDisable(GL_BLEND);

  if(!CheckForError())
    return false;

  m_transparency = 0.0f;

  return true;
}

bool GLLib::EnableTransparency(float t)
{
  assert(t >= 0.0f && t <= 1.0f);

  m_transparency = t;

  if(!CheckForError())
    return false;

  glEnable(GL_BLEND);      
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  if(!CheckForError())
    return false;

  return true;
}

bool GLLib::SetModelMatrix(GLLIB_matrix &matrix)
{
  if(!CheckForError())
    return false;

  // NOTE: Camera must be set before call to SetModelMatrix!

  glMatrixMode(GL_MODELVIEW);

  glPopMatrix();  // get can matrix
  glPushMatrix(); // duplicate cam matrix

  if(!CheckForError())
    return false;

  glMultMatrixf(matrix.GetPtr());
  if(!CheckForError())
    return false;

  return true;
}

bool GLLib::EnableDepthBuffer()
{
  if(!CheckForError())
    return false;

  glEnable(GL_DEPTH_TEST);

  if(!CheckForError())
    return false;

  return true;
}

bool GLLib::DisableDepthBuffer()
{
  if(!CheckForError())
    return false;

  glDisable(GL_DEPTH_TEST);

  if(!CheckForError())
    return false;

  return true;
}