#include "stdafx.h"
#include "glmodel.h"

#include "3dsloader.h"

#define X_MAXFLOAT  1000000000.0f
#define X_MINFLOAT -1000000000.0f

GLModel::GLModel(GLLib *gllib)
{
  assert(gllib != NULL);
  m_gllib = gllib;

  m_vtx    = NULL;
  m_uv     = NULL;
  m_normal = NULL;
  m_faces  = NULL;

  m_nvtx   = 0;
  m_nfaces = 0;

  m_type     = 0;

  m_ambient   = 0x000000;
  m_diffuse   = 0x000000;
  m_specular  = 0x000000;
  m_shininess = 0.0f;

  int m_texture = -1;

  m_transparent = false;
  m_wireframe   = false;

  m_matrix.Identity();
}

GLModel::~GLModel()
{
  if(m_vtx != NULL)
  {
    delete []m_vtx;
    m_vtx = NULL;
  }

  if(m_uv != NULL)
  {
    delete []m_uv;
    m_uv = NULL;
  }

  if(m_normal != NULL)
  {
    delete []m_normal;
    m_normal = NULL;
  }

  if(m_faces != NULL)
  {
    delete []m_faces;
    m_faces = NULL;
  }
}

bool GLModel::SetNVtxFaces(int nvtx, int nfaces, int type)
{
  m_type = type;

  if(m_vtx != NULL)
  {
    delete []m_vtx;
    m_vtx = NULL;
  }

  if(m_uv != NULL)
  {
    delete []m_uv;
    m_uv = NULL;
  }

  if(m_normal != NULL)
  {
    delete []m_normal;
    m_normal = NULL;
  }

  if(m_faces != NULL)
  {
    delete []m_faces;
    m_faces = NULL;
  }

  m_nvtx = nvtx;
  m_nfaces = nfaces;

  m_vtx = new GLLIB_vtx[m_nvtx];

  if(m_vtx == NULL)
  {
    m_nvtx = 0;
    return false;
  }
  
  // only create array if uv needed
  if(type & GLMODEL_TEX_UV)
  {
    m_uv = new GLLIB_uvvtx[m_nvtx];

    if(m_uv == NULL)
    {
      m_nvtx = 0;
      return false;
    }
  }
  
  // only create array if normals needed
  if(type & GLMODEL_TEX_NORMAL)
  {
    m_normal = new GLLIB_vtx[m_nvtx];

    if(m_normal == NULL)
    {
      m_nvtx = 0;
      return false;
    }
  }
  
  m_faces = new GLMODEL_face[m_nfaces];

  if(m_faces == NULL)
  {
    m_nfaces = 0;
    return false;
  }

  return true;
}

void GLModel::CalcSize()
{
  m_min_coordinate.x = X_MAXFLOAT;
  m_min_coordinate.y = X_MAXFLOAT;
  m_min_coordinate.z = X_MAXFLOAT;
  m_max_coordinate.x = X_MINFLOAT;
  m_max_coordinate.y = X_MINFLOAT;
  m_max_coordinate.z = X_MINFLOAT;

  m_middle_coordinate.x = 0;
  m_middle_coordinate.y = 0;
  m_middle_coordinate.z = 0;

  for(int i=0; i<m_nvtx; i++)
  {
    m_middle_coordinate = AddVector(m_middle_coordinate, m_vtx[i]);

    if(m_vtx[i].x > m_max_coordinate.x)
      m_max_coordinate.x = m_vtx[i].x;
    if(m_vtx[i].y > m_max_coordinate.y)
      m_max_coordinate.y = m_vtx[i].y;
    if(m_vtx[i].z > m_max_coordinate.z)
      m_max_coordinate.z = m_vtx[i].z;

    if(m_vtx[i].x < m_min_coordinate.x)
      m_min_coordinate.x = m_vtx[i].x;
    if(m_vtx[i].y < m_min_coordinate.y)
      m_min_coordinate.y = m_vtx[i].y;
    if(m_vtx[i].z < m_min_coordinate.z)
      m_min_coordinate.z = m_vtx[i].z;
  }

  m_middle_coordinate = MultiplyVector(m_middle_coordinate, 1/(float)m_nvtx);

  m_size = SubVector(m_max_coordinate, m_min_coordinate);
}

bool GLModel::CalcNormals()
{
  int i;

  assert(m_type & GLMODEL_TEX_NORMAL);

  if(m_type & GLMODEL_TEX_NORMAL == 0)
    return false;

  if(m_normal == NULL)
    return false;

  for(i=0; i<m_nvtx; i++)
  {
    m_normal[i] = InitVector();
  }

  for(i=0; i<m_nfaces; i++)
  {
    GLLIB_vtx a = SubVector(m_vtx[m_faces[i].p[1]], m_vtx[m_faces[i].p[0]]);
    GLLIB_vtx b = SubVector(m_vtx[m_faces[i].p[2]], m_vtx[m_faces[i].p[0]]);
    GLLIB_vtx n = CrossProduct(a, b);

    m_normal[m_faces[i].p[0]] = AddVector(m_normal[m_faces[i].p[0]], n);
    m_normal[m_faces[i].p[1]] = AddVector(m_normal[m_faces[i].p[1]], n);
    m_normal[m_faces[i].p[2]] = AddVector(m_normal[m_faces[i].p[2]], n);
  }

  for(i=0; i<m_nvtx; i++)
  {
    m_normal[i] = NormalizeVector(m_normal[i]);
  }

  return true;
}

bool GLModel::Render()
{
  if(m_vtx == NULL || m_faces == NULL)
    return false;

  m_gllib->SetModelMatrix(m_matrix);

  if(m_transparent)
    m_gllib->EnableTransparency(0.5f);

  if(m_wireframe)
  {
    m_gllib->DisableTexture(0);
    m_gllib->SetMaterial(0xFFFF00, 0, 0, 0.0f);

    m_gllib->RenderFacesWireframe(m_vtx, (int *)m_faces, m_nfaces);
  }
  else
  {
    if(m_ambient != 0x000000 || m_diffuse != 0x000000 || m_specular != 0x000000)
      m_gllib->SetMaterial(m_ambient, m_diffuse, m_specular, m_shininess);

    if(m_texture >= 0)
      m_gllib->SetCurrentTexture(0, m_texture);
    else
      m_gllib->DisableTexture(0);

    m_gllib->RenderFaces(m_vtx, m_uv, m_normal, (int *)m_faces, m_nfaces);
  }  

  if(m_transparent)
    m_gllib->DisableTransparency();

  return true;
}

bool GLModel::Load3DS(const char *filename, int object, float scale)
{
  int i;
  C3DSLoader c3ds;

  if(!c3ds.Load3DS(filename))
    return false;

  if(object < 0 || object >= (int)c3ds.m_nObjects)
    return false;

  load3ds_object_t *obj = &c3ds.m_pObjects[object];

  m_nfaces = obj->nfaces;
  m_faces = new GLMODEL_face[m_nfaces];

  m_nvtx = obj->nvertices;
  m_vtx = new GLLIB_vtx[m_nvtx];

  m_normal = new GLLIB_vtx[m_nvtx];

  m_uv = new GLLIB_uvvtx[m_nvtx];

  for(i=0; i<m_nvtx; i++)
  {
    m_vtx[i].x = -obj->vertices[i].vlocal.x*scale;
    m_vtx[i].y = obj->vertices[i].vlocal.y*scale;
    m_vtx[i].z = obj->vertices[i].vlocal.z*scale;

    if(obj->vertices[i].uv_set)
    {
      m_uv[i].u = obj->vertices[i].uv.u;
      m_uv[i].v = obj->vertices[i].uv.v;
    }
    else
    {
      m_uv[i].u = m_vtx[i].x/100;
      m_uv[i].v = m_vtx[i].z/100;
    }
  }

  for(i=0; i<m_nfaces; i++)
  {
    for(int j=0; j<3; j++)
    {
      m_faces[i].p[j] = obj->faces[i].v_index[j];

      assert(m_faces[i].p[j] >= 0 && m_faces[i].p[j] < m_nvtx);
    }
  }

  m_type = GLMODEL_TEX_UV | GLMODEL_TEX_NORMAL;

  m_matrix.Identity();

  CalcNormals();

  CalcSize();

  return true;
}

void GLModel::SetMaterial(GLLIB_COLOR ambient, GLLIB_COLOR diffuse, GLLIB_COLOR specular, float shininess)
{
  m_ambient = ambient;
  m_diffuse = diffuse;
  m_specular = specular;
  m_shininess = shininess;
}