#include "stdafx.h"
#include <math.h>
#include "mainengine.h"
#include "simplemainengine.h"
#include "bmploader.h"
#include "procrustes.h"

// parameters:
#define MAX_STEER   0.6f
#define STEER_SPEED 1.5f

#define SPEED_SCALE 1.0f

// camera
#define CAMDELAY_POS    0.4f
#define CAMDELAY_LOOKAT 0.4f
#define CAMDELAY_UP     0.4f

// particles
#define MAX_PARTICLES   1024

#define SCALE_SPRING    60

#define CAR_V0 1244     // vertices of 3D car model to attach chassis to
#define CAR_V1 884
#define CAR_V2 1099
#define CAR_V3 738

#define SPR_T0 (1+136)  // chassis vertices in spring model
#define SPR_T1 (2+136) 
#define SPR_T2 (3+136) 
#define SPR_T3 (0+136) 

#define MAX_MS_PER_FRAME 250

#define FILE3DS_CAR   "chevy_nowheels_3c.3ds"
#define FILE3DS_WHEEL "wheel.3ds"

SimpleMainEngine::SimpleMainEngine()
{
  m_carmodel = NULL;
  m_car_metal = NULL;
  m_car_window = NULL;
  m_car_seats = NULL;
  m_car_steer = NULL;

  m_wheelmodel = NULL;
  m_wheelinside = NULL;

  m_floormodel = NULL;

  m_particlemodel = NULL;

  m_sim = NULL;
  m_particle = NULL;

  m_cam = CAM_FIXED;

  m_rendercar = true;
  m_rendersprings = false;
  m_renderfloor = true;
  m_renderparticles = true;

  m_lookat_dir = LOOKAT_FRONT;

  m_simulated_time = 0.0f;
}

SimpleMainEngine::~SimpleMainEngine()
{
  if(m_carmodel != NULL)
  { 
    delete m_carmodel;
    m_carmodel = NULL;
  } 
 
  if(m_car_metal != NULL)
  {
    delete m_car_metal;
    m_car_metal = NULL;
  }

  if(m_car_window != NULL)
  {
    delete m_car_window;
    m_car_window = NULL;
  }

  if(m_car_seats != NULL)
  {
    delete m_car_seats;
    m_car_seats = NULL;
  }

  if(m_car_steer != NULL)
  {
    delete m_car_steer;
    m_car_steer = NULL;
  }

  if(m_wheelmodel != NULL) 
  { 
    delete m_wheelmodel; 
    m_wheelmodel = NULL; 
  } 
 
  if(m_wheelinside != NULL) 
  { 
    delete m_wheelinside;
    m_wheelinside = NULL; 
  } 
 
  if(m_floormodel != NULL) 
  { 
    delete m_floormodel; 
    m_floormodel = NULL; 
  } 

  if(m_particlemodel != NULL) 
  { 
    delete m_particlemodel; 
    m_particlemodel = NULL; 
  } 

  if(m_particle != NULL)
  {
    delete m_particle;
    m_particle = NULL;
  }

  if(m_sim != NULL)
  {
    delete m_sim;
    m_sim = NULL;
  }
}

void SimpleMainEngine::RenderSpringModel()
{
  int i;
  GLLIB_matrix initmatrix; 

  int nvtx = m_sim->getNumOfPts(); 
  GLLIB_vtx *points = m_sim->getPointAddress(); 

  int nsprings = m_sim->getNumOfSprings(); 
  Spring *springs = m_sim->getSpringAddress(); 

  m_gllib->SetModelMatrix(initmatrix); 

  m_gllib->SetMaterial(0xff0000, 0x00000, 0x000000, 0.0f);
  for(i=0; i<nvtx; i++)
  {
    GLLIB_vtx v; 
 
    v = MultiplyVector(points[i], SCALE_SPRING);
    m_gllib->RenderPoint(&v, 10.0f);
  } 
  
  m_gllib->SetMaterial(0xffffff, 0xffffff, 0xffffff, 0.0f);
  for(i=0; i<nsprings; i++) 
  { 
    GLLIB_vtx p1, p2; 

    p1 = MultiplyVector(points[springs[i].p1], SCALE_SPRING);
    p2 = MultiplyVector(points[springs[i].p2], SCALE_SPRING);
 
    m_gllib->RenderLine(&p1, &p2); 
  } 
}

GLLIB_matrix SimpleMainEngine::FitCar()
{
  GLLIB_matrix fitmatrix; 
  GLLIB_vtx from_vtx[4]; 
  GLLIB_vtx to_vtx[4]; 

  int nvtx = m_sim->getNumOfPts(); 
  GLLIB_vtx *points = m_sim->getPointAddress(); 
 
  // find trafo matrix of car 
  from_vtx[0] = m_carmodel->m_vtx[CAR_V0];
  from_vtx[1] = m_carmodel->m_vtx[CAR_V1];
  from_vtx[2] = m_carmodel->m_vtx[CAR_V2];
  from_vtx[3] = m_carmodel->m_vtx[CAR_V3];
 
  to_vtx[0] = MultiplyVector(points[SPR_T0], SCALE_SPRING);
  to_vtx[1] = MultiplyVector(points[SPR_T1], SCALE_SPRING);
  to_vtx[2] = MultiplyVector(points[SPR_T2], SCALE_SPRING);
  to_vtx[3] = MultiplyVector(points[SPR_T3], SCALE_SPRING);
  
  Procrustes::FitObject(fitmatrix, to_vtx, from_vtx, 4);

  return fitmatrix;
}

GLLIB_matrix SimpleMainEngine::FitWheel(int wheelindex)
{
  GLLIB_matrix fitmatrix;
  GLLIB_vtx from_vtx[4]; 
  GLLIB_vtx to_vtx[4]; 

  int nvtx = m_sim->getNumOfPts(); 
  GLLIB_vtx *points = m_sim->getPointAddress(); 
 
  int   AX1[4]    = {1, 35, 68, 102}; // face indices for Modell.txt !
  int   AX2[4]    = {0, 34, 69, 103};
  int   AXUP[4]   = {2, 36, 70, 104};
  float MIRROR[4] = {-1, -1, 1, 1};
 
  GLLIB_vtx wheel_axis1  = MultiplyVector(points[AX1[wheelindex]], SCALE_SPRING);
  GLLIB_vtx wheel_axis2  = MultiplyVector(points[AX2[wheelindex]], SCALE_SPRING);
  GLLIB_vtx wheel_axisup = MultiplyVector(points[AXUP[wheelindex]], SCALE_SPRING); 
 
  to_vtx[0] = wheel_axis1;
  to_vtx[1] = wheel_axis2;
  to_vtx[2] = wheel_axisup;
 
  from_vtx[0] = m_wheelmodel->m_middle_coordinate;
  from_vtx[0].z += m_wheelmodel->m_size.z/2;
  from_vtx[1] = m_wheelmodel->m_middle_coordinate;
  from_vtx[1].z -= m_wheelmodel->m_size.z/2;
  from_vtx[2] = m_wheelmodel->m_middle_coordinate;
  from_vtx[2].x += m_wheelmodel->m_size.x/2;
  from_vtx[2].z += MIRROR[wheelindex] * m_wheelmodel->m_size.z/2;
 
  Procrustes::FitObject(fitmatrix, to_vtx, from_vtx, 3); 
 
  return fitmatrix;
}

void SimpleMainEngine::CalculateStep(float time)
{
  int nvtx = m_sim->getNumOfPts(); 
  GLLIB_vtx *points = m_sim->getPointAddress(); 

  int nsprings = m_sim->getNumOfSprings(); 
  Spring *springs = m_sim->getSpringAddress(); 

  int nwheels = m_sim->getNumOfWheels(); 
  PoweredWheel *wheels = m_sim->getWheelAddress(); 

  float tosimulate = time - m_simulated_time;

  if(tosimulate > MAX_MS_PER_FRAME/1000.0f)
    tosimulate = MAX_MS_PER_FRAME/1000.0f;

  float angle = m_steervar.Step(tosimulate, m_key_state[ENGINE_KEY_RIGHT], m_key_state[ENGINE_KEY_LEFT]);
  
  m_sim->getSteeringAngleAddress()[0].setAngle(angle);
  m_sim->getSteeringAngleAddress()[1].setAngle(-angle);
  
  float power = m_throttlevar.Step(tosimulate, m_key_state[ENGINE_KEY_UP], m_key_state[ENGINE_KEY_DOWN]);
    
  m_eng.setThrottle(power);
   
  float pressure = m_breakvar.Step(tosimulate, m_key_state[ENGINE_KEY_SPACE], false);
    
  for(int i=0; i<nwheels; i++)
  {
    wheels[i].SetBreak(pressure); 
  }  

  // particles
  m_particle->Simulate(time, tosimulate);
  m_particle->SetTime(time);

  m_simulated_time += m_sim->step((int)(tosimulate*1000)) / 1000.0f;

  // how much time left to simulate?
  tosimulate = time - m_simulated_time;

  // too much -> skip time (-> not realtime anymore!)
  if(tosimulate > MAX_MS_PER_FRAME/1000.0f)
    m_simulated_time = time;
}

GLLIB_vtx SimpleMainEngine::SetCamera(float time)
{
  GLLIB_matrix fitmatrix;
  bool want_relative_campos = false;
  
  fitmatrix = FitCar();
  
  GLLIB_vtx campos;
  GLLIB_vtx camlookat;
  GLLIB_vtx camup = {0,1,0};

  switch(m_cam)
  {
    case CAM_FIXED:
      campos.x = 0;
      campos.y = 200;
      campos.z = 300;
      camlookat = fitmatrix.ApplyToVtx(m_carmodel->m_middle_coordinate);
      break;

    case CAM_STEER:
      {
        GLLIB_vtx cam_relative = {-20, 40, 40};
        GLLIB_vtx camlookat_relative = {-20,0,-200};
        GLLIB_vtx camup_relative = {0,1,0};

        switch(m_lookat_dir)
        {
          case LOOKAT_FRONT:
            camlookat_relative.x = -20;
            camlookat_relative.z = -200;
            break;
          case LOOKAT_LEFT:
            camlookat_relative.x = -200;
            camlookat_relative.z = 38;
            cam_relative.x = 20;
            cam_relative.z = 0;
            break;
          case LOOKAT_RIGHT:
            camlookat_relative.x = 200;
            camlookat_relative.z = 38;
            cam_relative.z = 0;
            break;
          case LOOKAT_BACK:
            camlookat_relative.x = -21;
            camlookat_relative.z = 200;
            cam_relative.z = 0;
            break;

          default:
            assert(0);
            break;
        }

        campos = fitmatrix.ApplyToVtx(cam_relative);
        camlookat = fitmatrix.ApplyToVtx(camlookat_relative);
        camup = fitmatrix.ApplyToVector(camup_relative);

        want_relative_campos = true;
      }
      break;

    case CAM_MOVIE:
      {
        GLLIB_vtx cam_relative = {-100, 0, 180};
        GLLIB_vtx camlookat_relative = {-100,0,-200};
        GLLIB_vtx camup_relative = {0,1,0};
        campos = fitmatrix.ApplyToVtx(cam_relative);
        camlookat = fitmatrix.ApplyToVtx(camlookat_relative);
        camup = fitmatrix.ApplyToVector(camup_relative);

        want_relative_campos = true;
      }
      break;

    case CAM_FOLLOW:
      {
        camlookat = fitmatrix.ApplyToVtx(m_carmodel->m_middle_coordinate);
        GLLIB_vtx cam_relative = {0,0,0};

        float h = 140.0f;
        float r = 600.0f;        

        switch(m_lookat_dir)
        {
          case LOOKAT_FRONT:
            cam_relative.x = 0;            
            cam_relative.z = r;
            break;
          case LOOKAT_LEFT:
            cam_relative.x = r;            
            cam_relative.z = 0;
            break;
          case LOOKAT_RIGHT:
            cam_relative.x = -r;            
            cam_relative.z = 0;
            break;
          case LOOKAT_BACK:
            cam_relative.x = 0;            
            cam_relative.z = -r;
            break;

          default:
            assert(0);
            break;
        }

        campos = fitmatrix.ApplyToVtx(cam_relative);
        campos.y = camlookat.y + h;
      }
      break;

    default:
      assert(false);
      break;
  }

  GLLIB_matrix inv = fitmatrix;
  inv.Inverse();


  GLLIB_vtx campos_relative = inv.ApplyToVtx(campos);
  m_campos_rel_interp.Set(campos_relative, time);
  m_campos_interp.Set(campos, time);

  GLLIB_vtx camlookat_relative = inv.ApplyToVtx(camlookat);
  m_camlookat_rel_interp.Set(camlookat_relative, time);
  m_camlookat_interp.Set(camlookat, time);

  GLLIB_vtx camup_relative = inv.ApplyToVector(camup);
  m_camup_rel_interp.Set(camup_relative, time);
  m_camup_interp.Set(camup, time);

  if(want_relative_campos)
  {
    campos = m_campos_rel_interp.Get(time, CAMDELAY_POS);
    campos = fitmatrix.ApplyToVtx(campos);

    camlookat = m_camlookat_rel_interp.Get(time, CAMDELAY_LOOKAT);
    camlookat = fitmatrix.ApplyToVtx(camlookat);

    camup = m_camup_rel_interp.Get(time, CAMDELAY_UP);
    camup = fitmatrix.ApplyToVector(camup);
  }
  else
  {
    campos = m_campos_interp.Get(time, CAMDELAY_POS);

    camlookat = m_camlookat_interp.Get(time, CAMDELAY_LOOKAT);    

    camup = m_camup_interp.Get(time, CAMDELAY_UP);
  }  

  m_gllib->SetCamera(campos, camlookat, camup);

  return campos;
}

void SimpleMainEngine::RenderSkyBox(GLLIB_vtx campos)
{
  GLLIB_uvvtx uv[4];
  GLLIB_uvvtx fuv[3];
  GLLIB_vtx   fv[3];
  float skybox_size = 3000;

  GLLIB_vtx p0[6] = {{-1,1,1}, {-1,1,-1}, {1,1,-1}, {1,1,1}, {-1,1,-1}, {-1,-1,1}};
  GLLIB_vtx p1[6] = {{1,1,1}, {-1,1,1}, {-1,1,-1}, {1,1,-1}, {1,1,-1}, {1,-1,1}};
  GLLIB_vtx p2[6] = {{1,-1,1}, {-1,-1,1}, {-1,-1,-1}, {1,-1,-1}, {1,1,1}, {1,-1,-1}};
  GLLIB_vtx p3[6] = {{-1,-1,1}, {-1,-1,-1}, {1,-1,-1}, {1,-1,1}, {-1,1,1}, {-1,-1,-1}};

  float epsilon = 0.001f;

  uv[0].u = epsilon;
  uv[0].v = epsilon;
  uv[1].u = 1-epsilon;
  uv[1].v = epsilon;
  uv[2].u = 1-epsilon;
  uv[2].v = 1-epsilon;
  uv[3].u = epsilon;
  uv[3].v = 1-epsilon;

  GLLIB_matrix m;
  m.Translate(campos);

  m_gllib->SetModelMatrix(m);

  m_gllib->DisableDepthBuffer();

  for(int i=0; i<6; i++)
  {
    m_gllib->SetCurrentTexture(0, m_skytex[i]);
    m_gllib->SetMaterial(0xFFFFFF, 0, 0, 0);

    fv[0] = MultiplyVector(p0[i], skybox_size);
    fv[1] = MultiplyVector(p1[i], skybox_size);
    fv[2] = MultiplyVector(p2[i], skybox_size);

    fuv[0] = uv[0]; fuv[1] = uv[1]; fuv[2] = uv[2];

    m_gllib->RenderFace(fv, fuv, NULL);

    fv[0] = MultiplyVector(p0[i], skybox_size);
    fv[1] = MultiplyVector(p2[i], skybox_size);
    fv[2] = MultiplyVector(p3[i], skybox_size);

    fuv[0] = uv[0]; fuv[1] = uv[2]; fuv[2] = uv[3];

    m_gllib->RenderFace(fv, fuv, NULL);
  }

  m_gllib->EnableDepthBuffer();
}

void SimpleMainEngine::RenderParticles(float time)
{
  int n_totalpart;
  int part_indexstart;
  int part_n;
  particle_t *part = m_particle->GetParticles(&n_totalpart, &part_indexstart, &part_n);

  for(int i=0; i<part_n; i++)
  {
    int j = (i + part_indexstart) % n_totalpart;

    if(part[j].alive)
    {
      float alpha = 1 - (time-part[j].life_start) / MAX_PARTICLE_LIFETIME;
      m_gllib->EnableTransparency(alpha);
      GLLIB_vtx scaledp;
      scaledp.x = part[j].p.x * SCALE_SPRING;
      scaledp.y = part[j].p.y * SCALE_SPRING;
      scaledp.z = part[j].p.z * SCALE_SPRING;

      m_particlemodel->m_matrix.Identity();
      m_particlemodel->m_matrix.Scale(part[j].rndvect);
      m_particlemodel->m_matrix.Scale(alpha*3 + 1);
      GLLIB_vtx rotp = {0,0,0};
      m_particlemodel->m_matrix.Rotate(rotp, part[j].rndvect, 0.0f*(rand()%1000)/1000.0f + j*0.7f + (1-alpha)*(1-alpha)*4);
      m_particlemodel->m_matrix.Translate(scaledp);

      m_particlemodel->Render();
    }
  }

  m_gllib->DisableTransparency();
}

void SimpleMainEngine::ThickLine(GLLIB_vtx v1, GLLIB_vtx v2, float d)
{
  GLLIB_vtx t[3];

  t[0] = v1;
  t[1] = v1; t[1].z += d;
  t[2] = v2;
  m_gllib->RenderFace(t, NULL, NULL);

  t[0] = v2;
  t[1] = v1; t[1].z += d;
  t[2] = v1;
  m_gllib->RenderFace(t, NULL, NULL);

  t[0] = v1;t[0].z += d;
  t[1] = v2;t[1].z += d;
  t[2] = v2;
  m_gllib->RenderFace(t, NULL, NULL);

  t[0] = v2;
  t[1] = v2;t[1].z += d;
  t[2] = v1;t[2].z += d;
  m_gllib->RenderFace(t, NULL, NULL);


  t[0] = v1;
  t[1] = v1; t[1].x += d;
  t[2] = v2;
  m_gllib->RenderFace(t, NULL, NULL);

  t[0] = v2;
  t[1] = v1; t[1].x += d;
  t[2] = v1;
  m_gllib->RenderFace(t, NULL, NULL);

  t[0] = v1;t[0].x += d;
  t[1] = v2;t[1].x += d;
  t[2] = v2;
  m_gllib->RenderFace(t, NULL, NULL);

  t[0] = v2;
  t[1] = v2;t[1].x += d;
  t[2] = v1;t[2].x += d;
  m_gllib->RenderFace(t, NULL, NULL);
}

bool SimpleMainEngine::RenderStep(float time)
{ 
  GLLIB_matrix fitmatrix;
  int i;

  // scale speed
  time *= SPEED_SCALE;

  // simulation step
  CalculateStep(time);
  
  m_gllib->StartRender(0x505070, false);
 
  // place camera
  GLLIB_vtx campos = SetCamera(time);

  RenderSkyBox(campos);
 
  // render floor 
  m_floormodel->Render();

  if(!m_renderfloor)
  {
    m_floormodel->m_wireframe = true;
    m_floormodel->Render();
    m_floormodel->m_wireframe = false;
  }

  if(m_rendersprings)
  {
    // redner springs'n'masses
    RenderSpringModel();
  }

  if(m_rendercar)
  {
    GLLIB_matrix wheelmatrix[4];
    // render wheels
    for(i=0; i<4; i++)
    {
      wheelmatrix[i] = FitWheel(i);

      m_wheelmodel->m_matrix.SetMatrix(wheelmatrix[i].GetPtr()); 
      m_wheelmodel->Render(); 

      m_wheelinside->m_matrix.SetMatrix(wheelmatrix[i].GetPtr());    
      m_wheelinside->Render();
    }  

    // render car
    fitmatrix = FitCar();

    m_carmodel->m_matrix.SetMatrix(fitmatrix.GetPtr());
    m_carmodel->Render(); 

    m_car_metal->m_matrix.SetMatrix(fitmatrix.GetPtr()); 
    m_car_metal->Render();
 
    m_car_seats->m_matrix.SetMatrix(fitmatrix.GetPtr()); 
    m_car_seats->Render();
 
    m_car_steer->m_matrix.Identity();

    GLLIB_vtx steeraxisrot = {0,0,1};
    GLLIB_vtx steeraxispos = {0,-10,0};
    GLLIB_vtx axispos = {0,0,0};
    GLLIB_vtx rotaxis = {1,0,0};
    GLLIB_vtx move = {-20,-20,-130};

    m_car_steer->m_matrix.Rotate(steeraxispos, steeraxisrot, -m_steervar.GetValue() * 4.0f);
    m_car_steer->m_matrix.Rotate(axispos, rotaxis, -0.3f);
    m_car_steer->m_matrix.Translate(move);
  
    m_car_steer->m_matrix.MultiplyByLeft(fitmatrix); 
    m_car_steer->Render();

    // render 'Federung'
    
    GLLIB_vtx *points = m_sim->getPointAddress(); 

    GLLIB_matrix initmatrix;
    m_gllib->SetModelMatrix(initmatrix);
 
    GLLIB_vtx center = m_carmodel->m_middle_coordinate;
    center.y -= 35;
    center = fitmatrix.ApplyToVtx(center);

    int model_pt[4]  = {CAR_V0, CAR_V1, CAR_V2, CAR_V3};
    int spring_pt[4] = {35, 68, 1, 102};
    m_gllib->SetMaterial(0x000000, 0, 0, 0.0f);

    for(int j=0; j<4; j++)
    {
      GLLIB_vtx p = m_carmodel->m_vtx[model_pt[j]];
      p.z *= 0.95f;
      GLLIB_vtx v1 = fitmatrix.ApplyToVtx(p);
      GLLIB_vtx v2 = MultiplyVector(points[spring_pt[j]], SCALE_SPRING);

      ThickLine(v1, v2, 5.0f);
      ThickLine(center, v2, 5.0f);
    }
 
    // note: render front window last because it's transparent!
    m_car_window->m_matrix.SetMatrix(fitmatrix.GetPtr()); 
    m_car_window->Render();
  }

  if(m_renderparticles)
  {
    RenderParticles(time);
  }

  m_gllib->EndRender();

  return true;
}

bool SimpleMainEngine::ResetSim()
{
  if(m_sim != NULL)
  {
    delete m_sim;
    m_sim = NULL;
  }

  m_sim = new SpringSim;

  // load physical model
  if(!m_sim->init("Modell.txt")) 
  {
    assert(0);
    return false;
  }

  int nsprings = m_sim->getNumOfSprings(); 
  Spring *springs = m_sim->getSpringAddress(); 

  m_steervar.Init(0, -MAX_STEER, MAX_STEER, STEER_SPEED, STEER_SPEED);

  m_throttlevar.Init(0, -1.0f, 1.0f, 1.0f/0.5f, 1.0f/0.5f);
  m_breakvar.Init(0, -1.0f, 1.0f, 1.0f/0.5f, 1.0f/0.5f);

  m_eng.init(m_sim->getWheelAddress());

  // initialize particles
  if(m_particle != NULL)
  {
    delete m_particle;
    m_particle = NULL;
  }

  m_particle = new ParticleSim(MAX_PARTICLES, m_sim->cd.getGround());
  m_sim->cd.SetParticles(m_particle);

  return true;
}

bool SimpleMainEngine::InitAll()
{
  int i;

  // create floor texture
  BmpLoader bmp;
  if(!bmp.LoadFromFile("isampler48.bmp")) // NOTE: must be bmp
  {
    // texture not found or wrong format
    assert(0);
    return false;
  }
  bmp.ResizeToTexture();
  m_floortex = m_gllib->CreateTexture(bmp.GetRAWptr(), bmp.GetXRes(), bmp.GetYRes());

  // create metal texture
  if(!bmp.LoadFromFile("gray_structure.bmp")) // NOTE: must be bmp
  {
    // texture not found or wrong format
    assert(0);
    return false;
  }
  bmp.ResizeToTexture();
  m_metaltex = m_gllib->CreateTexture(bmp.GetRAWptr(), bmp.GetXRes(), bmp.GetYRes());

  // create leather texture
  if(!bmp.LoadFromFile("leather.bmp")) // NOTE: must be bmp
  {
    // texture not found or wrong format
    assert(0);
    return false;
  }
  bmp.ResizeToTexture();
  m_leathertex = m_gllib->CreateTexture(bmp.GetRAWptr(), bmp.GetXRes(), bmp.GetYRes());

  // create skybox textures
  char skyboxname[6][20] = {"desert_rt.bmp", "desert_bk.bmp", "desert_lf.bmp", 
                            "desert_ft.bmp", "desert_up.bmp", "desert_dn.bmp"};
  for(i=0; i<6; i++)
  {
    if(!bmp.LoadFromFile(skyboxname[i])) // NOTE: must be bmp
    {
      // texture not found or wrong format
      assert(0);
      return false;
    }
    bmp.ResizeToTexture();
    m_skytex[i] = m_gllib->CreateTexture(bmp.GetRAWptr(), bmp.GetXRes(), bmp.GetYRes());
  }

  // create car model
  m_carmodel = new GLModel(m_gllib); 
  if(!m_carmodel->Load3DS(FILE3DS_CAR, 2, 20.0f)) 
  { 
    assert(0); 
    return false; 
  } 
  m_carmodel->SetMaterial(0xa07070, 0xa04040, 0xffffff, 0.8f);
  m_carmodel->m_texture = m_metaltex;
  
  m_car_metal = new GLModel(m_gllib);
  if(!m_car_metal->Load3DS(FILE3DS_CAR, 3, 20.0f)) 
  { 
    assert(0); 
    return false; 
  } 
  m_car_metal->SetMaterial(0x606060, 0x707070, 0xf0f0f0, 1.0f);
  m_car_metal->m_texture = m_metaltex;
  
  m_car_seats = new GLModel(m_gllib);
  if(!m_car_seats->Load3DS(FILE3DS_CAR, 1, 20.0f)) 
  { 
    assert(0); 
    return false; 
  } 
  m_car_seats->SetMaterial(0xa0a0a0, 0x707070, 0x202020, 0.2f);
  m_car_seats->m_texture = m_leathertex;
  
  m_car_steer = new GLModel(m_gllib);
  if(!m_car_steer->Load3DS(FILE3DS_CAR, 4, 20.0f)) 
  { 
    assert(0); 
    return false; 
  } 
  m_car_steer->SetMaterial(0x909090, 0xa0a0a0, 0xf0f0f0, 0.5f);
  
  m_car_window = new GLModel(m_gllib);
  if(!m_car_window->Load3DS(FILE3DS_CAR, 0, 20.0f)) 
  { 
    assert(0); 
    return false; 
  } 
  m_car_window->SetMaterial(0x5050a0, 0x5050e0, 0xe0e0e0, 0.8f);
  m_car_window->m_transparent = true;
  
  if(!ResetSim())
    return false;
 
  // create wheel 
  m_wheelmodel = new GLModel(m_gllib);  
  if(!m_wheelmodel->Load3DS(FILE3DS_WHEEL, 1, 3.0f)) 
  { 
    assert(0); 
    return false; 
  } 
  m_wheelmodel->SetMaterial(0x303030, 0x707070, 0x0a0a0a, 0.3f);
  m_wheelmodel->m_texture = m_floortex;
  
  // create wheel (metal parts)
  m_wheelinside = new GLModel(m_gllib);  
  if(!m_wheelinside->Load3DS(FILE3DS_WHEEL, 0, 3.0f)) 
  {
    assert(0);
    return false;
  }
  m_wheelinside->SetMaterial(0x404040, 0x606060, 0xffffff, 0.9f);

  // create particle model
  m_particlemodel = new GLModel(m_gllib);
  m_particlemodel->SetNVtxFaces(4, 4, GLMODEL_TEX_UV);
  m_particlemodel->SetMaterial(0x707070, 0xe0e0e0, 0x000000, 0.2f);
  m_particlemodel->m_texture = m_floortex;

  float PART_SCALE = 4.0f; // particle scale

  m_particlemodel->m_vtx[0].x = 0.0f * PART_SCALE;
  m_particlemodel->m_vtx[0].y = 0.0f * PART_SCALE;
  m_particlemodel->m_vtx[0].z = 0.0f * PART_SCALE;
  m_particlemodel->m_vtx[1].x = 1.0f * PART_SCALE;
  m_particlemodel->m_vtx[1].y = 0.0f * PART_SCALE;
  m_particlemodel->m_vtx[1].z = 0.0f * PART_SCALE;
  m_particlemodel->m_vtx[2].x = 0.5f * PART_SCALE;
  m_particlemodel->m_vtx[2].y = 1.0f * PART_SCALE;
  m_particlemodel->m_vtx[2].z = 0.0f * PART_SCALE;
  m_particlemodel->m_vtx[3].x = 0.5f * PART_SCALE;
  m_particlemodel->m_vtx[3].y = 0.5f * PART_SCALE;
  m_particlemodel->m_vtx[3].z = 1.0f * PART_SCALE;

  m_particlemodel->m_uv[0].u = 0.0f;
  m_particlemodel->m_uv[0].v = 0.0f;
  m_particlemodel->m_uv[1].u = 0.05f;
  m_particlemodel->m_uv[1].v = 0.0f;
  m_particlemodel->m_uv[2].u = 0.0f;
  m_particlemodel->m_uv[2].v = 0.06f;
  m_particlemodel->m_uv[3].u = 0.07f;
  m_particlemodel->m_uv[3].v = 0.04f;

  m_particlemodel->m_faces[0].p[0] = 0;
  m_particlemodel->m_faces[0].p[1] = 1;
  m_particlemodel->m_faces[0].p[2] = 2;
  m_particlemodel->m_faces[1].p[0] = 1;
  m_particlemodel->m_faces[1].p[1] = 3;
  m_particlemodel->m_faces[1].p[2] = 2;
  m_particlemodel->m_faces[2].p[0] = 3;
  m_particlemodel->m_faces[2].p[1] = 0;
  m_particlemodel->m_faces[2].p[2] = 2;
  m_particlemodel->m_faces[3].p[0] = 3;
  m_particlemodel->m_faces[3].p[1] = 1;
  m_particlemodel->m_faces[3].p[2] = 0;
   
  // create floor
  m_floormodel = new GLModel(m_gllib);
  m_floormodel->SetNVtxFaces((GROUND_W+1)*(GROUND_W+1), GROUND_W*GROUND_W*2, GLMODEL_TEX_UV | GLMODEL_TEX_NORMAL);
  m_floormodel->SetMaterial(0x707070, 0xf0f0f0, 0x000000, 0.4f);
  m_floormodel->m_texture = m_floortex;

  Ground *ground = m_sim->cd.getGround();

  GLLIB_matrix initmatrix;
  for(i=0; i<GROUND_W+1; i++)
  {
    for(int j=0; j<GROUND_W+1; j++)
    {
      GLLIB_vtx v = ground->GetVtx(i, j);
      v = MultiplyVector(v, SCALE_SPRING);

      m_floormodel->m_vtx[j*(GROUND_W+1)+i] = v;      

#define TMPC 0.3f
      m_floormodel->m_uv[j*(GROUND_W+1)+i].u = i*TMPC;
      m_floormodel->m_uv[j*(GROUND_W+1)+i].v = j*TMPC;
    }
  }

  for(i=0; i<GROUND_W; i++)
  {
    for(int j=0; j<GROUND_W; j++)
    {
      m_floormodel->m_faces[(j*GROUND_W+i)*2+0].p[0] = j*(GROUND_W+1) + i;
      m_floormodel->m_faces[(j*GROUND_W+i)*2+0].p[1] = (j+1)*(GROUND_W+1) + (i+1);
      m_floormodel->m_faces[(j*GROUND_W+i)*2+0].p[2] = j*(GROUND_W+1) + (i+1);

      m_floormodel->m_faces[(j*GROUND_W+i)*2+1].p[0] = j*(GROUND_W+1) + i;
      m_floormodel->m_faces[(j*GROUND_W+i)*2+1].p[1] = (j+1)*(GROUND_W+1) + i;
      m_floormodel->m_faces[(j*GROUND_W+i)*2+1].p[2] = (j+1)*(GROUND_W+1) + (i+1);
    }
  }
 
  m_floormodel->CalcNormals(); 

  // reset key states

  for(i=0; i<ENGINE_MAX_KEYS; i++)
    m_key_state[i] = false;

  return true;
}

void SimpleMainEngine::SetKeyState(int key, bool state)
{
  assert(key >= 0 && key < ENGINE_MAX_KEYS);

  if(key < 0 || key >= ENGINE_MAX_KEYS)
    return;

  m_key_state[key] = state;

  if(state)
  {
    switch(key)
    {
      case ENGINE_KEY_F1:
        m_cam = CAM_FIXED;
        break;
      case ENGINE_KEY_F2:
        m_cam = CAM_STEER;
        break;
      case ENGINE_KEY_F3:
        m_cam = CAM_MOVIE;
        break;
      case ENGINE_KEY_F4:
        m_cam = CAM_FOLLOW;
        break;
      case ENGINE_KEY_F5:
        m_rendercar = !m_rendercar;
        break;
      case ENGINE_KEY_F6:
        m_rendersprings = !m_rendersprings;
        break;
      case ENGINE_KEY_F7:
        m_renderparticles = !m_renderparticles;
        break;
      case ENGINE_KEY_F8:
        m_renderfloor = !m_renderfloor;
        break;
      case ENGINE_KEY_ESC:
        ResetSim();
        break;
      default:
        break;
    }
  }

  bool lookat_left = (m_lookat_dir == LOOKAT_LEFT) || (m_lookat_dir == LOOKAT_BACK);
  bool lookat_right = (m_lookat_dir == LOOKAT_RIGHT) || (m_lookat_dir == LOOKAT_BACK);

  if(key == 'q')
    lookat_left = state;

  if(key == 'e')
    lookat_right = state;

  if(lookat_left && lookat_right)
    m_lookat_dir = LOOKAT_BACK;
  else
  if(lookat_left)
    m_lookat_dir = LOOKAT_LEFT;
  else
  if(lookat_right)
    m_lookat_dir = LOOKAT_RIGHT;
  else
    m_lookat_dir = LOOKAT_FRONT;
}
